// -*- c++ -*-
//
// Copyright 1993 Massachusetts Institute of Technology
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of M.I.T. not be used in advertising or
// publicity pertaining to distribution of the software without specific,
// written prior permission.  M.I.T. makes no representations about the
// suitability of this software for any purpose.  It is provided "as is"
// without express or implied warranty.
//
// M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
// ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
// M.I.T.  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
// ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER IN AN ACTION * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// Copyright (c) 1991, 1992 Redwood Design Automation
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the name of
// Redwood Design Automation may not be used in any advertising or publicity
// relating to the software without the specific, prior written permission of
// Redwood Design Automation.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// Author: Denis Bohm  Sep. 1992 
// Revised: Joerg Ringel Nov. 1992 
// Further abuse: Dave Sternlicht Feb. 1993

// debug mode
//#define TREE_DEBUG
// 72 dots per inch.
#define DPI 72.0                        
// for allocating largest array 
#define TREE_INITIAL_DEPTH 10		

#include <assert.h>

#include <InterViews/canvas.h>
#include <InterViews/event.h>
#include <InterViews/hit.h>
#include <InterViews/patch.h>
#include <InterViews/resource.h>
#include <InterViews/window.h>
#include <OS/list.h>
#include <OS/math.h>
#include "tree.h"

const Coord global_brush_width = 1.0;
static const float epsilon = 1e-4;

//
// TreeInfo Declaration
//

class TreeInfo {
public:
  enum {Requested = 0x01, Allocated = 0x02, Resized = 0x04, Hidden = 0x08};
  
  TreeInfo();
  TreeInfo(Glyph* glyph);
  ~TreeInfo();
  
  void finalize();
  void replace(Glyph* glyph);
  void span(Coord& w, Coord& h);

  Glyph* glyph_;
  GlyphIndex index_;
  Coord x_, y_, w_, h_;
  int status_;
  char* name_;
  Requisition requisition_;
  Allocation allocation_;
  Extension extension_;

  // constraint information
  TreeInfo* tree_root_;             // tested by constrain_append 
  TreeInfo* top_node_;
  TreeInfo* parent_;
  TreeInfo** children_;            // array of ptrs to TreeInfo's
  int n_children_, max_children_;
  Coord bbsubwidth_, bbsubheight_; // bounding box of sub tree.
  Coord bbwidth_, bbheight_;       // bounding box including node.
  Coord cx_, cy_;
  Patch* p_;                       // call realloc on it to get a requirement.

private:
  void std_init();
};

//
// TreeImpl Declaration
//

class TreeImpl : public Glyph, public Adjustable {
public:
  TreeImpl(Session*);
  virtual ~TreeImpl();  
  void layout_tree(boolean inset_values);
  void constraint_append(Glyph*, Glyph*);
  
  // Glyph
  virtual void request(Requisition&) const;
  virtual void allocate(Canvas*, const Allocation&, Extension&);
  virtual void draw(Canvas*, const Allocation&) const;
  virtual void pick(Canvas*, const Allocation&, int depth, Hit&);
  virtual void undraw();
  virtual void append(Glyph*);
  virtual void prepend(Glyph*);
  virtual void insert(GlyphIndex, Glyph*);
  virtual void remove(GlyphIndex);
  virtual void replace(GlyphIndex, Glyph*);
  virtual void change(GlyphIndex);
  virtual GlyphIndex count() const;
  virtual Glyph* component(GlyphIndex) const;
  virtual void allotment(GlyphIndex, DimensionName, Allotment&) const;

  // Adjustable
  virtual Coord upper(DimensionName) const;
  virtual Coord cur_lower(DimensionName) const;
  virtual Coord cur_upper(DimensionName) const;
  virtual Coord cur_length(DimensionName) const;
  virtual void scroll_to(DimensionName, Coord lower);
  virtual void scale_to(DimensionName, float fraction_visible);
  virtual void zoom_to(float magnification);

private:
  // Internal Tree Manager Functions
  void insert_node(TreeInfo* parent, TreeInfo* child);
  void layout_node_lines(TreeInfo*);

  void initialize_dimensions(Coord** listp, int* sizep, int n);
  void insert_node();
  void delete_node();
  
  void compute_bounding_box_subtree(TreeInfo*, int depth);
  void set_positions(TreeInfo*, int level);
  void arrange_subtree(TreeInfo*, int depth, Coord x, Coord y);
  void set_tree_size(boolean inset_values, Coord, Coord);

  // Line Functions
  void num_lines(int);
  int num_lines();

  //
  // Methods that support panning.
  //

  // like Page (but not a page because Tree can pan and zoom) 
  virtual Coord left() const;
  virtual Coord right() const;
  virtual Coord bottom() const;
  virtual Coord top() const;
  virtual Coord x() const;
  virtual Coord y() const;

  virtual GlyphIndex index(const Glyph* glyph) const;

  virtual void move(GlyphIndex, Coord x, Coord y);
  virtual void location(GlyphIndex, Coord& x, Coord& y);

  virtual void size(GlyphIndex, Coord width, Coord height);
  virtual Coord width(GlyphIndex) const;
  virtual Coord height(GlyphIndex) const;
  virtual void natural(GlyphIndex, boolean natural);
  virtual boolean natural(GlyphIndex) const;

  virtual void depth(GlyphIndex, unsigned long depth);
  virtual unsigned long depth(GlyphIndex) const;

  virtual void show(GlyphIndex, boolean);
  virtual boolean showing(GlyphIndex) const;

  // like Patch (but not a patch because Tree is not a MonoGlyph) 
  virtual Canvas* canvas() const;
  virtual const Transformer& transformer() const;
  virtual const Allocation& allocation() const;
  virtual const Extension& extension() const;

  virtual void redraw() const;
  virtual void reallocate();
  virtual void repick(int depth, Hit&);

  virtual void damage(const Extension&) const;
  virtual void updateMatrix();
  virtual void check(GlyphIndex);
  virtual boolean inside(const Allocation& a) const;
  virtual void check(Allocation a) const;
  virtual void repair() const;
  virtual boolean boundingBox() const;
  virtual boolean defined() const;
  virtual Coord bbLeft() const;
  virtual Coord bbBottom() const;
  virtual Coord bbRight() const;
  virtual Coord bbTop() const;

  void invariants() const;

  //
  // Internal Tree Data
  //

  // The TreePart nested class contains fields similar
  // to the Xaw tree widgets class record.
  struct TreePart {
    Coord hpad;                 // horizontal space between children.
    Coord vpad;                 // vertical space between clhidren.
    TreeInfo* tree_root;        // hidden root of all children. 
    Coord* largest;             // list of largest per depth.
    int n_largest;              // number of elements in largest.
    Coord maxwidth, maxheight;  // for shrink wrapping.
  } tpart_;

  Canvas* canvas_;
  Session* session_;
  WidgetKit* kit_;
  TreeInfoPtrList* info_;
  TreeDepthList* depth_;
  Coord tree_width_, tree_height_; // similar to Xt core width & height.
  int num_child_nodes_;
  boolean defined_:1;
  Transformer matrix_;
  Allocation allocation_;
  Extension extension_;
  int num_lines_;                  // number of lines
  Coord l_, b_, r_, t_, cl_, cb_, cr_, ct_;
  Coord sw_, sh_, vw_, vh_;
  int line_point_counter_;
  Coord dx_, dy_;                  // translation displacment
  Coord* xpoints_;                 // coordinates describing the button tree.
  Coord* ypoints_;  
};

//
// TreeInfo Methods
//

TreeInfo::TreeInfo() {
  glyph_ = nil;
  std_init();
}

TreeInfo::TreeInfo(Glyph* glyph) {
  glyph_ = glyph;
  Resource::ref(glyph_);
  std_init();
}

TreeInfo::~TreeInfo() { Resource::unref(glyph_); }

void TreeInfo::std_init() {
  // lets be object oriented :-/
  x_ = y_ = w_ = h_ = 0.0;
  status_ = n_children_ = max_children_ = n_children_ = max_children_ = 0;
  tree_root_ = top_node_ = parent_ = nil; children_ = nil; p_ = nil;
  bbsubwidth_ = bbsubheight_ = bbwidth_ = bbheight_ = 0.0;
  cx_ = cy_ = 0;
}

void TreeInfo::finalize() {
  glyph_->undraw();
  Resource::unref(glyph_);
  glyph_ = nil;
}

void TreeInfo::replace(Glyph* glyph) {
  Resource::ref(glyph);
  glyph_->undraw();
  Resource::unref(glyph_);
  glyph_ = glyph;
  status_ = 0;
}

void TreeInfo::span(Coord& w, Coord& h) {
  // cache the requisition.
  if (! (status_ & Requested)) {
    glyph_->request(requisition_);
    status_ |= Requested;
  }

  // calculate the new allocation.
  Requirement& sx = requisition_.x_requirement();
  Requirement& sy = requisition_.y_requirement();
  if (status_ & Resized) {
    w = Math::min(Math::max(w_, sx.natural() - sx.shrink()), sx.natural() +
      sx.stretch());
    h = Math::min(Math::max(h_, sy.natural() - sy.shrink()), sy.natural() +
      sy.stretch());
  } else {
    w = sx.natural();
    h = sy.natural();
  }
}

//
// TreeInfo List 
//
declarePtrList(TreeInfoPtrList,TreeInfo);
implementPtrList(TreeInfoPtrList,TreeInfo); 

//
// TreeDepth List 
//
declareList(TreeDepthList,GlyphIndex);
implementList(TreeDepthList,GlyphIndex);

//
// Tree Methods
//

Tree::Tree(Session* s) {
  timpl_ = new TreeImpl(s); 
  timpl_->attach(Dimension_X, this); 
  timpl_->attach(Dimension_Y, this); 
}

Tree::~Tree() { delete timpl_; }

void Tree::force_layout(Orientation) { 
  // LeftToRight Orientation nyi
  timpl_->layout_tree(True); 
}

void Tree::insert_node(Glyph* parent, Glyph* child) { 
  timpl_->constraint_append(parent, child);
}

// Glyph
void Tree::request(Requisition& r) const { timpl_->request(r); }

void Tree::allocate(Canvas* c, const Allocation& a, Extension& e) {
  timpl_->allocate(c,a,e);
}

void Tree::draw(Canvas* c, const Allocation& a) const {
  timpl_->draw(c, a);
}

void Tree::pick(Canvas* c, const Allocation& a, int depth, Hit& h) {
  timpl_->pick(c, a, depth, h);
}

void Tree::allotment(GlyphIndex i, DimensionName d, Allotment& a) const {
  timpl_->allotment(i, d, a);
}

// Adjustable
Coord Tree::upper(DimensionName d) const  { return timpl_->upper(d); };

Coord Tree::cur_lower(DimensionName d) const { return timpl_->cur_lower(d); };

Coord Tree::cur_upper(DimensionName d) const { return timpl_->cur_upper(d); };

Coord Tree::cur_length(DimensionName d) const { return timpl_->cur_length(d); }

void Tree::scroll_to(DimensionName d, Coord lower)  { 
  timpl_->scroll_to(d, lower); 
}

void Tree::scale_to(DimensionName d, float fraction_visible) {
  timpl_->scale_to(d, fraction_visible);
}

void Tree::zoom_to(float magnification) {
  timpl_->zoom_to(magnification);
}

// Observer
void Tree::update(Observable*) { notify_all(); }

//
// TreeImpl Methods
//

void TreeImpl::invariants() const {
  assert(depth_->count() == info_->count());
  long count = depth_->count();
  for (long i = 0; i < count; ++i) {
    assert((0 <= depth_->item(i)) && (depth_->item(i) < count));
  }
}

TreeImpl::TreeImpl(Session* s) {
  session_ = s;                    // Session not a glyph to be ref'ed/unref'ed
  info_ = new TreeInfoPtrList;
  depth_ = new TreeDepthList;
  defined_ = false;
  num_child_nodes_ = 0;
  canvas_ = nil;
  small_scroll(Dimension_X, 10.0);
  small_scroll(Dimension_Y, 10.0);
  l_ = b_ = cl_ = cb_ = 0.0; r_ = 10.0*DPI; t_ = 10.0*DPI;
  cr_ = 5.0*DPI; ct_ = 5.0*DPI;
  sw_ = 10.0*DPI; sh_ = 10.0*DPI; vw_ = 5.0*DPI; vh_ = 5.0*DPI;
  num_lines(0);
  kit_ = WidgetKit::instance();

  // create the hidden root glyph...
  constraint_append(nil, nil);

  tpart_.largest = nil;
  tpart_.n_largest = 0;
  tpart_.hpad = 20.0;
  tpart_.vpad = 6.0;

  initialize_dimensions(&tpart_.largest, &tpart_.n_largest,TREE_INITIAL_DEPTH);
}

TreeImpl::~TreeImpl() {
  GlyphIndex count = info_->count();
  for (GlyphIndex index = 0; index < count; ++index) {
    TreeInfo* info = info_->item(index);
    info->finalize();
  }
  delete info_;
  delete depth_;
}

//
// Page like
//

Coord TreeImpl::left() const { return allocation_.left(); }

Coord TreeImpl::right() const { return allocation_.right(); }

Coord TreeImpl::bottom() const { return allocation_.bottom(); }

Coord TreeImpl::top() const { return allocation_.top(); }

Coord TreeImpl::x() const { return allocation_.x(); }

Coord TreeImpl::y() const { return allocation_.y(); }

GlyphIndex TreeImpl::index(const Glyph* glyph) const {
  GlyphIndex count = info_->count();
  for (GlyphIndex index = 0; index < count; ++index) {
    if (info_->item(index)->glyph_ == glyph) {
      return index;
    }
  }
  return -1; // error condition
}

void TreeImpl::move(GlyphIndex index, Coord x, Coord y) {
  TreeInfo* info = info_->item(index);
  info->x_ = x;
  info->y_ = y;
}

void TreeImpl::location(GlyphIndex index, Coord& x, Coord& y) {
  TreeInfo* info = info_->item(index);
  x = info->x_;
  y = info->y_;
}

void TreeImpl::size(GlyphIndex index, Coord width, Coord height) {
  TreeInfo* info = info_->item(index);
  info->status_ |= TreeInfo::Resized;
  info->w_ = width;
  info->h_ = height;
}

Coord TreeImpl::width(GlyphIndex index) const {
  TreeInfo* info = info_->item(index);
  return info->w_;
}

Coord TreeImpl::height(GlyphIndex index) const {
  TreeInfo* info = info_->item(index);
  return info->h_;
}

void TreeImpl::natural(GlyphIndex index, boolean natural) {
  TreeInfo* info = info_->item(index);
  if (natural) {
    info->status_ &= ~TreeInfo::Resized;
  } else {
    info->status_ |= TreeInfo::Resized;
  }
}

boolean TreeImpl::natural(GlyphIndex index) const {
  TreeInfo* info = info_->item(index);
  return ! (info->status_ & TreeInfo::Resized);
}

void TreeImpl::depth(GlyphIndex index, unsigned long depth) {
  long count = depth_->count();
  for (long i = 0; (i < count) && (depth_->item(i) != index); ++i) {
  }
  depth_->remove(i);
  depth_->insert(count - 1 - depth, index);
  TreeInfo* info = info_->item(index);
  if (info->status_ & TreeInfo::Allocated) {
    damage(info->extension_);
  }
  invariants();
}

unsigned long TreeImpl::depth(GlyphIndex index) const {
  long count = depth_->count();
  for (long i = 0; (i < count) && (depth_->item(i) != index); ++i) {
  }
  return count - 1 - i;
}

void TreeImpl::show(GlyphIndex index, boolean showing) {
  TreeInfo* info = info_->item(index);
  if (((info->status_ & TreeInfo::Hidden) == 0) != showing) {
    damage(info->extension_);
    if (showing) {
      info->status_ &= ~TreeInfo::Hidden;
    } else {
      info->status_ |= TreeInfo::Hidden;
    }
  }
}

boolean TreeImpl::showing(GlyphIndex index) const {
  return (info_->item(index)->status_ & TreeInfo::Hidden) == 0;
}

//
// Patch like
//

Canvas* TreeImpl::canvas() const { return canvas_; }

const Transformer& TreeImpl::transformer() const { return matrix_; }

const Allocation& TreeImpl::allocation() const { return allocation_; }

const Extension& TreeImpl::extension() const { return extension_; }

void TreeImpl::redraw() const {
  if (canvas_) {
    canvas_->damage(extension_);
  }
}

void TreeImpl::reallocate() {
  if (canvas_) {
    Extension extension;
    allocate(canvas_, allocation_, extension);
  }
}

void TreeImpl::repick(int depth, Hit& hit) {
  if (canvas_) {
    pick(canvas_, allocation_, depth, hit);
  }
}

//
// Glyph
//

void TreeImpl::request(Requisition& requisition) const {

  Coord w = 1.0;
  Coord h = 1.0;

  if (defined()) {
    w = bbRight() - bbLeft() + 1.0;
    h = bbTop() - bbBottom() + 1.0;
  }

  w *= cr_;
  h *= ct_;

  requisition.require(Dimension_X, Requirement(w, fil, w, 0));
  requisition.require(Dimension_Y, Requirement(h, fil, h, 0));
}

void TreeImpl::allocate(Canvas* c, const Allocation& a, Extension& ext) {
  boundingBox();

  c->push_transform();
  c->transform(matrix_);

  dx_ = a.left() / cr_ - cl_;
  dy_ = a.bottom() / ct_ - cb_;

  GlyphIndex count = info_->count();
  for (GlyphIndex index = 0; index < count; ++index) {
    TreeInfo* info = info_->item(index);

    if (! info->glyph_) {
      continue;
    }

    // calculate the new allocation.
    Coord w, h;
    info->span(w, h);
    Requirement& sx = info->requisition_.x_requirement();
    Requirement& sy = info->requisition_.y_requirement();
    Allotment nax(dx_ + info->x_, w, sx.alignment());
    Allotment nay(dy_ + info->y_, h, sy.alignment());

    // check to see of the new allocation is different than the old.
    Allocation& a = info->allocation_;
    Allotment& ax = a.x_allotment();
    Allotment& ay = a.y_allotment();
    if ((! (info->status_ & TreeInfo::Allocated)) ||
      (! ax.equals(nax, epsilon)) || (! ay.equals(nay, epsilon))) {

      // damage the old extension.
      Extension& b = info->extension_;
      if (info->status_ & TreeInfo::Allocated) {
        damage(b);
      }

      // reallocate the glyph.
      b.clear();
      a.allot_x(nax);
      a.allot_y(nay);
      info->glyph_->allocate(c, a, b);

      // damage the new extension.
      damage(b);
    }
    info->status_ |= TreeInfo::Allocated;
  }

  c->pop_transform();
  ext.merge(c, a);

  canvas_ = c;
  allocation_ = a;
  extension_ = ext;
}

void TreeImpl::draw(Canvas* c, const Allocation& a) const {
  // draw lines according to the ctm:
  const Color* black = Color::lookup(session_->default_display(), "black");
  Brush* br = new Brush(global_brush_width);

  c->clip_rect(a.left(), a.bottom(), a.right(), a.top());

  long count = depth_->count();
  for (long i = 0; i < count; ++i) {
    GlyphIndex index = depth_->item(i);
    TreeInfo* info = info_->item(index);
    if (info->glyph_ && (info->status_ & TreeInfo::Allocated) &&
      (! (info->status_ & TreeInfo::Hidden))) {
      Extension& b = info->extension_;
      if (c->damaged(b)) {
        Allocation& a = info->allocation_;
        info->glyph_->draw(c, a);
      }
    }
  }

  Coord tx1, ty1, tx2, ty2;

  Transformer t;
  t.translate(dx_, dy_);

  int index = 0;
  for (int line_cnt = 0; line_cnt < num_lines_; line_cnt++) {
    c->new_path();
    t.transform(xpoints_[index], ypoints_[index], tx1, ty1);  
    t.transform(xpoints_[index + 1], ypoints_[index + 1], tx2, ty2);
    c->move_to(tx1, ty1);
    c->line_to(tx2, ty2);
    index += 2;
    c->close_path();
    c->stroke(black, br);
  }
}

void TreeImpl::pick(Canvas* c, const Allocation&, int depth, Hit& h) {
  c->push_transform();
  c->transform(matrix_);
  h.push_transform();
  h.transform(matrix_);
  Coord left = h.left();
  Coord bottom = h.bottom();
  Coord right = h.right();
  Coord top = h.top();

  long count = depth_->count();
  for (long i = 0; i < count; ++i) {
    GlyphIndex index = depth_->item(i);
    TreeInfo* info = info_->item(index);
    if (info->glyph_ && (info->status_ & TreeInfo::Allocated) &&
      (! (info->status_ & TreeInfo::Hidden))) {
      Allocation& a = info->allocation_;
      if ((right >= a.left()) && (left <= a.right()) &&
        (top >= a.bottom()) && (bottom <= a.top())) {
        h.begin(depth, this, index);
        info->glyph_->pick(c, a, depth + 1, h);
        h.end();
      }
    }
  }

  h.pop_transform();
  c->pop_transform();
}

void TreeImpl::undraw() {
  GlyphIndex count = info_->count();
  for (GlyphIndex index = 0; index < count; ++index) {
    TreeInfo* info = info_->item(index);
    if (info->glyph_) {
      info->glyph_->undraw();
      info->status_ &= ~TreeInfo::Allocated;
    }
  }
  canvas_ = nil;
}

void TreeImpl::append(Glyph* glyph) {
  info_->append(new TreeInfo(glyph));

  GlyphIndex index = info_->count() - 1;
  depth_->append(index);
  info_->item(0)->index_ = 0;
  check(info_->count() - 1);
  invariants();
}

void TreeImpl::constraint_append(Glyph* parent, Glyph* child) {
  TreeInfo* pti = nil;

  if (parent) {                                    // child node
    // find the parents TreeInfo node:
    GlyphIndex count = info_->count();
    for (GlyphIndex index = 0; index < count; ++index) 
      if (info_->item(index)->glyph_ == parent)
        pti = info_->item(index);
  }

  if (child) {
    TreeInfo* tip = new TreeInfo;
    num_child_nodes_++;
    tip->glyph_ = child;
    tip->parent_ = pti;  
    tip->tree_root_ = tpart_.tree_root;

    info_->append(tip); // causes a TreeInfo deallocation

    GlyphIndex index = tip->index_ = info_->count() - 1;
    depth_->append(index);
    info_->item(0)->index_ = 0;
    check(info_->count() - 1);
    invariants();
  
    // If this glyph has a super-node, add it to that glyphs sub-nodes list.
    // Otherwise make it a sub-node of the tree_root glyph.
    if (parent)
      insert_node(tip->parent_, tip);
    else if (tip->tree_root_)
      insert_node(tpart_.tree_root, tip);
    return;
  }

    if (!parent && !child) {                       // root node
      tpart_.tree_root = new TreeInfo;
      tpart_.tree_root->w_ = 1.0;
      tpart_.tree_root->h_ = 1.0;
    }
}

void TreeImpl::prepend(Glyph* glyph) {
  info_->prepend(new TreeInfo(glyph));

  long count = depth_->count();
  for (long i = 0; i < count; ++i) {
    GlyphIndex& in = depth_->item_ref(i);
    ++in;
  }

  GlyphIndex index = 0;
  depth_->append(index);

  check(0);
  invariants();
}

void TreeImpl::insert(GlyphIndex index, Glyph* glyph) {
  info_->insert(index, new TreeInfo(glyph));

  long count = depth_->count();
  for (long i = 0; i < count; ++i) {
    GlyphIndex& in = depth_->item_ref(i);
    if (in >= index) {
      ++in;
    }
  }
  depth_->append(index);

  check(index);
  invariants();
}

void TreeImpl::remove(GlyphIndex index) {
  TreeInfo* info = info_->item(index);
  if (info->status_ & TreeInfo::Allocated) {
    damage(info->extension_);
  }
  boolean allocated = info->status_ & TreeInfo::Allocated;
  Allocation a = info->allocation_;
  info->finalize();
  info_->remove(index);

  long count = depth_->count();
  for (long i = 0; (i < count) && (depth_->item(i) != index); ++i) {
  }
  depth_->remove(i);

  count = depth_->count();
  for (i = 0; i < count; ++i) {
    GlyphIndex& in = depth_->item_ref(i);
    if (in >= index) {
      --in;
    }
  }

  if (allocated) {
    check(a);
  }
  invariants();
}

void TreeImpl::replace(GlyphIndex index, Glyph* glyph) {
  TreeInfo* info = info_->item(index);
  if (info->status_ & TreeInfo::Allocated) {
    damage(info->extension_);
  }
  boolean allocated = info->status_ & TreeInfo::Allocated;
  Allocation a = info->allocation_;
  info->replace(glyph);
  check(index);
  if (allocated) {
    check(a);
  }
}

void TreeImpl::change(GlyphIndex index) {
  TreeInfo* info = info_->item(index);
  info->status_ &= ~TreeInfo::Requested;
  if (info->status_ & TreeInfo::Allocated) {
    check(info->allocation_);
  }
  check(index);
}

GlyphIndex TreeImpl::count() const { return info_->count(); }

Glyph* TreeImpl::component(GlyphIndex index) const {
  return info_->item(index)->glyph_;
}

void TreeImpl::allotment(GlyphIndex i, DimensionName res, Allotment& a) const {
  a = info_->item(i)->allocation_.allotment(res);
}

//
// Adjustable
//

Coord TreeImpl::upper(DimensionName d) const {
  return d == Dimension_X ? sw_ : sh_;
}

Coord TreeImpl::cur_lower(DimensionName d) const {
  return d == Dimension_X ? cl_ : cb_;
}

Coord TreeImpl::cur_upper(DimensionName d) const {
  return d == Dimension_X ? cr_ : ct_;  
}

Coord TreeImpl::cur_length(DimensionName d) const {
  return d == Dimension_X ? vw_ : vh_;
}

void TreeImpl::scroll_to(DimensionName d, Coord p) {
  Coord a = lower(d);
  Coord b = upper(d) - cur_length(d);
  if (p < a) {
    p = a;
  } else if (p > b) {
    p = b;
  }
  if (d == Dimension_X) { cl_ = p; cr_ = p + cur_length(d); }
  else { cb_ = p; ct_ = p + cur_length(d); }
  reallocate();
  redraw();
  notify(d);
}

void TreeImpl::scale_to(DimensionName dimension, float fraction_visible) {
  // !!! shouldn't need to do this! -denis
  GlyphIndex count = info_->count();
  for (GlyphIndex index = 0; index < count; ++index) {
    TreeInfo* info = info_->item(index);
    if (info->glyph_) {
      info->glyph_->undraw();
      info->status_ &= ~TreeInfo::Allocated;
    }
  }

  *(dimension == Dimension_X ? &cr_ : &ct_) =
    allocation_.allotment(dimension).span() /
    (length(dimension) * fraction_visible);
  updateMatrix();
  reallocate();
  redraw();
  notify(dimension);
}

void TreeImpl::zoom_to(float magnification) {
  // !!! shouldn't need to do this! -denis
  GlyphIndex count = info_->count();
  for (GlyphIndex index = 0; index < count; ++index) {
    TreeInfo* info = info_->item(index);
    if (info->glyph_) {
      info->glyph_->undraw();
      info->status_ &= ~TreeInfo::Allocated;
    }
  }

  cr_ = magnification;
  ct_ = magnification;
  updateMatrix();
  reallocate();
  redraw();
  notify(Dimension_X);
  notify(Dimension_Y);
}

void TreeImpl::damage(const Extension& e) const {
  if (canvas_) {
    Coord l = Math::max(e.left(), extension_.left());
    Coord b = Math::max(e.bottom(), extension_.bottom());
    Coord r = Math::min(e.right(), extension_.right());
    Coord t = Math::min(e.top(), extension_.top());
    if ((l < r) && (b < t)) {
      canvas_->damage(l, b, r, t);
    }
  }
}

void TreeImpl::updateMatrix() {
  matrix_ = Transformer();
  matrix_.scale(cr_, ct_);
}

void TreeImpl::check(GlyphIndex index) {
  TreeInfo* info = info_->item(index);
  if (info->glyph_) {
    Coord w, h;
    info->span(w, h);
    Requirement& sx = info->requisition_.x_requirement();
    Requirement& sy = info->requisition_.y_requirement();
    Allotment ax(info->x_, w, sx.alignment());
    Allotment ay(info->y_, h, sy.alignment());
    Allocation a;
    a.allot_x(ax);
    a.allot_y(ay);
    check(a);
  }
}

boolean TreeImpl::inside(const Allocation& a) const {
  return (a.left() > l_) && (a.bottom() > b_) &&
    (a.right() < r_) && (a.top() < t_);
}

void TreeImpl::check(Allocation a) const {
  Coord dx = allocation_.left() - cl_;
  Coord dy = allocation_.bottom() - cb_;
  a.x_allotment().offset(-dx);
  a.y_allotment().offset(-dy);
  if (!inside(a)) boundingBox();
}

void TreeImpl::repair() const {
  if (canvas_) {
    Window* window = canvas_->window();
    if (window) window->repair();
  }
}

boolean TreeImpl::boundingBox() const {
  TreeImpl& self = *((TreeImpl*) this);

  boolean defined = false;
  Coord left, bottom, right, top;

  left = bottom = right = top = 0;
  GlyphIndex count = info_->count();
  for (GlyphIndex index = 0; index < count; ++index) {
    TreeInfo* info = info_->item(index);
    if (! info->glyph_) {
      continue;
    }
    
    // calculate the new allocation.
    Coord w, h;
    info->span(w, h);
    Requirement& sx = info->requisition_.x_requirement();
    Requirement& sy = info->requisition_.y_requirement();
    Allotment ax(info->x_, w, sx.alignment());
    Allotment ay(info->y_, h, sy.alignment());

    if (! defined) {
      left   = ax.begin();
      bottom = ay.begin();
      right  = ax.end();
      top    = ay.end();
      defined = true;
    } else {
      left   = Math::min(left, ax.begin());
      bottom = Math::min(bottom, ay.begin());
      right  = Math::max(right, ax.end());
      top    = Math::max(top, ay.end());
    }
  }

  boolean xChanged = (defined != defined_) || (left != l_) || (right != r_);
  boolean yChanged = (defined != defined_) || (bottom != b_) || (top != t_);

  self.defined_ = defined;
  self.l_   = left;
  self.b_ = bottom;
  self.r_  = right;
  self.t_    = top;

  if (xChanged || yChanged) {
    if (xChanged) notify(Dimension_X);
    if (yChanged) notify(Dimension_Y);
  }

  return xChanged || yChanged;
}

boolean TreeImpl::defined() const { return defined_; }

Coord TreeImpl::bbLeft() const { return l_; }

Coord TreeImpl::bbBottom() const { return b_; }

Coord TreeImpl::bbRight() const { return r_; }

Coord TreeImpl::bbTop() const { return t_; }

void TreeImpl::num_lines(int i) { num_lines_ = i; }

int TreeImpl::num_lines() { return num_lines_; }

Coord get_natural(TreeInfo* t, DimensionName d) {
    Coord result;
    Requisition r;
    t->glyph_->request(r);
    result = d == Dimension_X ? 
        r.x_requirement().natural() : r.y_requirement().natural();
    return result;
}

void TreeImpl::layout_node_lines(TreeInfo* tnode) {
  Coord parentx = tnode->cx_ + get_natural(tnode, Dimension_X);
  Coord parenty = tnode->cy_ + get_natural(tnode, Dimension_Y) / 2.0;
  for (int i = 0; i < tnode->n_children_; i++) {
    if (tnode->children_[i]) layout_node_lines(tnode->children_[i]);
    xpoints_[line_point_counter_] = parentx;
    ypoints_[line_point_counter_++] = parenty;
    xpoints_[line_point_counter_] = tnode->children_[i]->cx_;
    ypoints_[line_point_counter_++] = 
      tnode->children_[i]->cy_ + 
      get_natural(tnode->children_[i], Dimension_Y) / 2;
  }
}

void TreeImpl::insert_node(TreeInfo* parent, TreeInfo* node) {
  node->parent_ = parent;
  if (parent == nil) return;
  
  // If there isn't more room in the children array, 
  // allocate additional space.
  
  int nindex = parent->n_children_;
  
  if (parent->n_children_ == parent->max_children_) {
    parent->max_children_ += (parent->max_children_ / 2) + 2;
    
    typedef TreeInfo* TreeInfoPtr;
    
    parent->children_ = parent->children_ ?
      (TreeInfo**)realloc(
        (char*)parent->children_, 
	(unsigned int)parent->max_children_ * sizeof(TreeInfoPtr)
      ) :
      (TreeInfo**)malloc(
        (unsigned int)parent->max_children_ * sizeof(TreeInfoPtr)
      );
  } 

  // Add the sub_node in the next available slot and 
  // increment the counter.
  parent->children_[nindex] = node;
  parent->n_children_++;
}

void TreeImpl::initialize_dimensions(Coord** listp, int* sizep, int n) {
  if (!*listp) {
    *listp = new Coord[n];
    *sizep = ((*listp) ? n : 0);
    return;
  }
  if (n > *sizep) {
    realloc((char*)*listp, (unsigned int)(n*sizeof(Coord)));
    if (!*listp) {
      *sizep = 0;
      return;
    }
    *sizep = n;
  }
  return;
}

void TreeImpl::compute_bounding_box_subtree(TreeInfo* tinfo, int depth) {
  Coord newwidth, newheight;
  // Coord bw2 = tinfo->w_ * 2; glyphs don't have borrder width...
  
  if (tinfo->glyph_) {
    tinfo->p_ = new Patch(tinfo->glyph_);
    Requisition requis;
    tinfo->glyph_->request(requis);
    tinfo->w_ = requis.x_requirement().natural();
    tinfo->h_ = requis.y_requirement().natural();
    
  }
  
  // set the max-size per level.
  if (depth >= tpart_.n_largest) 
    initialize_dimensions (&tpart_.largest, &tpart_.n_largest, depth + 1);
  newwidth = tinfo->w_; // + bw2...
  if (tpart_.largest[depth] < newwidth)
    tpart_.largest[depth] = newwidth;
  
  // initialize
  tinfo->bbwidth_ = tinfo->w_;
  tinfo->bbheight_ = tinfo->h_;
#ifdef TREE_DEBUG    
  printf("compute_bounding_box#1: bbheight = %f\n", tinfo->bbheight_);
#endif
  if (tinfo->n_children_ == 0) return;
  
  // Figure the size of the opposite dimension (vertical if tree is 
  // horizontal, else vice versa).  The other dimension will be set 
  // in the second pass once we know the maximum dimensions.
  newwidth = newheight = 0.0;
  for (int i = 0; i < tinfo->n_children_; i++) {
    TreeInfo* child = tinfo->children_[i];
    
    compute_bounding_box_subtree(child, depth + 1);
    
    if (newwidth < child->bbwidth_) newwidth = child->bbwidth_;
    newheight += tpart_.vpad + child->bbheight_;
  }
  
  tinfo->bbsubwidth_ = newwidth;
  tinfo->bbsubheight_ = newheight;
  
  // Now fit parent onto side (or top) of bounding box and correct for
  // extra padding.  Be careful of unsigned arithmetic.
  tinfo->bbwidth_ += tpart_.hpad + newwidth;
  newheight -= tpart_.vpad;
  if (newheight > tinfo->bbheight_) tinfo->bbheight_ = newheight;
#ifdef TREE_DEBUG 
  printf("compute_bounding_box#2: bbheight = %f\n", tinfo->bbheight_);
#endif
}

void TreeImpl::set_positions(TreeInfo* tinfo, int level) {
  if (tinfo) {
    if (level > 0) {
      // move the glyph into position.
      
#ifdef TREE_DEBUG          
      printf(
	     "index_ = %d, cx_ = %f, cy_ = %f, name_ = %s\n", 
	     tinfo->index_, tinfo->cx_, tinfo->cy_, tinfo->name_
	     );
#endif
      move (tinfo->index_, tinfo->cx_, tinfo->cy_); 
    }
    
    // Set the positions of all children.
    for (int i = 0; i < tinfo->n_children_; i++) 
      set_positions(tinfo->children_[i], level + 1);
  }
};

void TreeImpl::arrange_subtree(TreeInfo* tinfo, int depth, Coord x, Coord y) {
  if (tinfo == nil) printf("ROOT NODE\n");
  // glyph ptr
  Glyph* gp = tinfo->glyph_;
  // tree info ref
  TreeInfo& tir = *tinfo;
  register int i;
    Coord newx, newy;
  TreeInfo* child = nil;
  Coord tmp;
  Coord bw2 = 0;
  boolean relayout = True;
  
  // If no children, then just lay out where requested.
#ifdef TREE_DEBUG 
  printf("arrange_subtree where req:, tir.cx_ = %f, tir.cy_ = %f\n", x, y);
#endif
  tir.cx_ = x;
  tir.cy_ = y;
  
  // only horizontal tree for now
  Coord myh = tir.h_;
  
  if (myh > tir.bbsubheight_) {
    y += (myh - tir.bbsubheight_) / 2;
    relayout = False;
#ifdef TREE_DEBUG 
    printf(
      "arrange_subtree: myh = %f, tir.bbsubheight_, %f, y = %f\n", 
      myh, tir.bbsubheight_, y
    );
#endif
  }

  if (tmp = (x + tir.bbwidth_) > tpart_.maxwidth)
    tpart_.maxwidth = tmp;
  
  if (tmp = (y + tir.bbheight_) > tpart_.maxheight)
    tpart_.maxheight = tmp;
  
  if (tir.n_children_ == 0) return;
  
  // Have children, so walk down tree laying out children, then laying
  // out parents.
  newx = x + tpart_.largest[depth];
  newy = y;
  if (depth > 0) newx += tpart_.hpad;
  
  for (i = 0; i < tir.n_children_; i++) {
    child = tir.children_[i];
    arrange_subtree (child, depth + 1, newx, newy);
#ifdef TREE_DEBUG 
    printf("arrange_subtree: vpad = %f\n" ,tpart_.vpad);
    printf("arrange_subtree: tinfo->bbheight_ = %f\n", tinfo->bbheight_);
#endif
    newy += tpart_.vpad + child->bbheight_;
  }
  
  // now layout parent between first and last children
  Coord adjusted;
  TreeInfo* firstcc = tinfo->children_[0];
  TreeInfo* lastcc = child;
  
  if (relayout) {
    // Adjustments are disallowed if they result in a position above
    // or to the left of the originally requested position, because
    // this could collide with the position of the previous sibling.
    tinfo->cx_ = x;
    adjusted = firstcc->cy_ + 
      ((lastcc->cy_ + child->h_ -
	firstcc->cy_ - tinfo->h_ + 1.0) / 2.0);
    if (adjusted > tinfo->cy_) tinfo->cy_ = adjusted;
#ifdef TREE_DEBUG 
    printf(
	   "arrange_subtree: (horiz), tinfo->cx_ = %f, tinfo->cy_ = %f\n", 
	   tinfo->cx_, tinfo->cy_
	   );
#endif
  }
};

// resizes the tree. changes the size of the panner control knob.
void TreeImpl::set_tree_size(boolean inset_values, Coord width, Coord height) {
  if (inset_values) {
    tree_width_ = width;
    tree_height_ = height;
  }
  // note there is no query of the parent to allow a resize...
};

void TreeImpl::layout_tree(boolean inset_values) {
  // Do a depth-first search computing the width and height of the bounding
  // box for the tree at that position (and below).  Then, walk again using
  // this information to layout the children at each level.
  if (tpart_.tree_root == nil) return;
  
  tpart_.maxwidth = tpart_.maxheight = 0;
  int i; Coord* dp;
  for (i = 0, dp = tpart_.largest; i < tpart_.n_largest; i++, dp++) *dp = 0;
  initialize_dimensions(
    &tpart_.largest, &tpart_.n_largest, tpart_.n_largest
  );
  compute_bounding_box_subtree(tpart_.tree_root, 0);
  
  // Second pass to do final layout.  Each child's bounding box is stacked
  // on top of (if horizontal, else next to) on top of its siblings.  The
  // parent is centered between the first and last children.
  arrange_subtree(tpart_.tree_root, 0, 0, 0);
  
  // Move each glyph into place.
  set_tree_size(inset_values, tpart_.maxwidth, tpart_.maxheight);
  set_positions(tpart_.tree_root, 0);
  
  // draw lines
  num_lines(num_child_nodes_ - 1);  // don't count "real" node.
  xpoints_ = new Coord[num_lines() * 2];
  ypoints_ = new Coord[num_lines() * 2];
  line_point_counter_ = 0;
  layout_node_lines(tpart_.tree_root->children_[0]);  
  
  // And redisplay.
  redraw();
  reallocate();
  redraw();
};

